#!/usr/bin/env python3

import os
import sys
import time
import math
import shlex
import signal
import argparse
import tempfile
import traceback
import urllib.parse
import subprocess as sp
from datetime import datetime

"""
x265ify.py: transcode video files to h265, third edition
2021-08-08, v1.4, ed <irc.rizon.net>, MIT-Licensed
https://ocv.me/dev/?x265ify.py

this replaces https://ocv.me/dev/?x265ify.bat

existing features carried over from the batch ver:
- CTRL-C once to terminate an encode cleanly

new features in this ver:
- makes the output folder if necessary
- support list of filenames on stdin
- support filenames starting with -
- support encoding from URLs with parameters
- support files with broken dts/pts
- open two MPVs for comparison in the cleanup-oneliner
"""

try:
    import ctypes
    from ctypes import wintypes

    _GetLongPathNameW = ctypes.windll.kernel32.GetLongPathNameW
    _GetLongPathNameW.argtypes = [wintypes.LPCWSTR, wintypes.LPWSTR, wintypes.DWORD]
    _GetLongPathNameW.restype = wintypes.DWORD
    
    WINDOWS = True
except:
    WINDOWS = False


CLEANUP = []


def tinput():
    try:
        input()
    except:
        pass


def longify(sPath):
    sz = 0
    for _ in range(64):
        buf = ctypes.create_unicode_buffer(sz)
        sz2 = _GetLongPathNameW(sPath, buf, sz)
        if sz2 == 0:
            return None
        if sz >= sz2:
            return buf.value
        
        sz = sz2

    raise Exception("uhh")


def need_shortnames():
    print("\nthis script requires the python drophandler to provide 8.3 names, so please switch from the WSH to the EXE handler (because WSH doesn't support unicode)")
    reg_path = os.path.abspath(sys.argv[0]) + ".reg"
    try:
        with open(reg_path, 'wb') as f:
            f.write(b"\xff\xfe" + r"""Windows Registry Editor Version 5.00

[HKEY_CLASSES_ROOT\Python.File\shellex\DropHandler]
@="{86C86720-42A0-1069-A2E8-08002B30309D}"

[HKEY_LOCAL_MACHINE\SOFTWARE\Classes\Python.File\shellex\DropHandler]
@="{86C86720-42A0-1069-A2E8-08002B30309D}"

; orig: wsh {60254CA5-953B-11CF-8C96-00AA00B8708C}
; works: exe {86C86720-42A0-1069-A2E8-08002B30309D}
""".encode('utf-16le'))

        print(f"\nthere's a registry fix you can apply here:\n    {reg_path}\n")
        print(f"hit ENTER to apply the fix, or CTRL-C to abort\n")
        input()
        sp.check_call(['regedit', reg_path], shell=True)
        print(f"\nok now you have to restart explorer.exe for that to take effect,\nor alternatively reassign the open-with for .py files\n")
        print(f"(hit ENTER to close this window)\n")

    except:
        print(f'\ntried (and failed) to create a registry fix at this path:\n    {reg_path}\n')


def sighandler(sig, frame):
    pass


def set_title(title):
    if WINDOWS:
        ctypes.windll.kernel32.SetConsoleTitleW(title)
    else:
        print(f'\033]0;{title}\033\\', end='')


def get_fps(fn):
    """
    don't rely on the rate/timebase in the container, it can be wrong:

    for f in *.mkv *.mp4 *.webm; do sleep 0.1;
    v1=$(ffprobe -v error -select_streams v:0 -show_entries stream=avg_frame_rate -of default=noprint_wrappers=1:nokey=1 "$f");
    v2=$(ffprobe -threads 0 -hide_banner -v fatal -select_streams v:0 -show_frames -show_entries frame=pkt_dts_time -of compact=p=0:nk=1 -read_intervals %+10 "$f" </dev/null | awk 'NR==0{t=$1;next} {f++} $1>=t+1{printf "%.3f\n", f/($1-t); t=$1; f=0; if (++p>=10) {exit}}' | sort | tail -n5 | head -n1);
    printf '%10s %6s \033[36m%s\033[0m\n' $v1 $v2 "$f"; done

    correct:
      30000/1001 29.970 Bitconnect haunted mine-a0UeUkKcU0U.mp4
      24000/1001 23.976 Doja Cat - Say So (Official Video)-pok8H_KF1FA.mkv
            30/1 30.000 Matsuoka Yoshitsugu with SAO harem-Qb8T0Qway5o.mkv
       19001/317 59.940 Playing A Link To The Past While Daydreaming About Magfest-tOXSkq0C5PE.mkv
    wrong:
    634200/21121 30.000 _-u2jzoo.mp4
      37440/1249 30.000 ed - railgun.mkv does it again (epilepsy warning maybe)-1233823315082534912.mp4
    2395000/79833 30.000 Dunkmaster - #NintendoSwitch-1238669448833187840.mp4
    185049000/6168301 30.000 KINOtheDJ 2020-03-28 03_27-37292108992.mp4
    """

    # using best_effort_timestamp_time (which should be pts)
    # because pkt_dts_time can be "N/A"
    txt = sp.check_output([
        "ffprobe",
        "-threads", "0",
        "-hide_banner",
        "-v", "fatal",
        "-select_streams", "v:0",
        "-show_frames",
        "-show_entries", "frame=best_effort_timestamp_time",
        "-read_intervals", "%+60",  # 1min
        fn])

    #print(txt.decode('utf-8'))
    lines = [float(x.split(b'=')[1]) for x in txt.replace(b'\r', b'').split(b'\n') if x.startswith(b"best_effort_timestamp_time=")]
    samples = []
    rtime = lines[0]
    frames = 1
    for ln in lines[1:]:
        if ln >= rtime + 1:
            samples.append(frames / (ln - rtime))
            rtime = ln
            frames = 0
        frames += 1

    # print('\n'.join([f'{x:.3f}' for x in samples]))
    samples.sort()
    ofs = int(len(samples) / 4)
    samples = samples[ofs:-ofs]
    try:
        return sum(samples) / len(samples)
    except:
        return 1


def get_res(fn):
    txt = sp.check_output([
        "ffprobe",
        "-hide_banner",
        "-v", "fatal",
        "-select_streams", "v:0",
        "-show_entries", "stream=width,height",
        fn])

    tab = {}
    lines = txt.replace(b'\r', b'').split(b'\n')
    for ln in lines:
        if b'=' in ln:
            k, v = ln.split(b'=')
            tab[k] = int(v)

    return tab[b'width'], tab[b'height']


def transcode(ar, fc_path, src, name_extra):
    preset = ar.p
    crf = ar.c
    ia = ar.ia
    oa = ar.oa
    vf = ar.vf

    if ar.r is not None:
        print(f'{src} fps = ...')
        try:
            src_fps = get_fps(src)
        except:
            src_fps = 0
        
        print(f'{src} fps = {src_fps:.3f}')
        
        if src_fps > ar.r:
            # note:
            #   `select` must precede `mpdecimate` (OK)
            #   but should be after denoisers (TODO?)
            div = int(math.ceil(src_fps / ar.r))
            vf = rf'select=not(mod(n-1\,{div})),{vf}'
            out_fps = src_fps / div

            print(f'{src} div = {div}')
            print(f'{src} fps = {out_fps:.3f}')

            name_extra += f'{round(out_fps):.0f}fps-'

    if ar.sz:
        src_w, src_h = get_res(src)
        print(f'{src} res = {src_w}x{src_h}')

        #print(f"{ar.sz} * 1.2 ({ ar.sz * 1.2 }) <= {src_w} ?")
        if ar.sz * 1.2 <= src_w:
            vf += f",scale={ar.sz}:-8"
            name_extra += f"{ar.sz}x-"

    # cfg: seek 9 seconds into video if "ID INVADED" in filename
    if 'ID INVADED' in src:
        ia += ' -ss 9'

    if 'x264' in preset:
        enc = 'h264'
        preset = 'veryslow'
    elif 'vp9' in preset:
        enc, preset, maxrate = preset.split('-')
        spd = preset.lstrip('s')
    else:
        enc = 'h265'

    # cfg: write output to ~/enc
    outdir = os.path.join(os.path.expanduser('~'), 'enc')
    try:
        os.mkdir(outdir)
    except:
        pass

    if '://' in src:
        src_name = src.split('/')[-1].split('?')[0]
        src_name = urllib.parse.unquote(src_name)
        src_mod = '8888-88-88 88:88:88'
        src_sz = 9001
        # ia += ' -reconnect 1 -reconnect_at_eof 1 -reconnect_streamed 1 -reconnect_delay_max 4294'
        # ^ doesnt even do anything haha i love computer
    else:
        src_name = os.path.basename(src)
        src_sz = os.path.getsize(src)
        src_mod = os.path.getmtime(src)
        src_mod = datetime.utcfromtimestamp(src_mod)
        src_mod = src_mod.strftime("%Y-%m-%d %H:%M:%S")

    outfile = os.path.join(outdir, src_name)
    #outfile += f".h265-720p-{preset}{crf}.mkv"
    outfile += f".{enc}-{name_extra}{preset}-{crf}.mkv"
    
    logfile = outfile + ".log"
    tmp_log = f"x265ify-{int(time.time()*1000)/1000.}-{os.getpid()}.log"
    tmp_log = os.path.join(os.path.expanduser('~'), tmp_log)
    ff_tmplog = tmp_log.replace('\\', '\\\\').replace(':', '\\:')

    if os.path.exists(outfile):
        return False

    lockdir = outfile + ".lock"
    try:
        os.mkdir(lockdir)
        CLEANUP.append(lockdir)
    except:
        return False

    try:
        os.unlink(logfile)
    except:
        pass

    env = os.environ.copy()
    env["FONTCONFIG_FILE"] = fc_path
    env["FFREPORT"] = f"level=32:file={ff_tmplog}"

    print(f"processing [{src}]")
    print(f"writing to [{outfile}]")
    print(f"logfile at [{tmp_log}]")
    print()

    cmd = [
        "ffmpeg",
        "-hide_banner",
        "-vsync", "vfr"
    ]
    cmd.extend(shlex.split(ia))
    cmd.extend([
        "-i", src,
        "-map", "0",
        "-map", "-0:d?",
        "-c", "copy",
        "-vf", vf
    ])

    if enc == 'h265':
        cmd.extend([
            "-c:v", "libx265",
            "-preset", preset
        ])
    elif enc == 'x264':
        cmd.extend([
            "-c:v", "libx264",
            "-preset", preset
        ])
    else:
        cmd.extend([
            "-c:v", "libvpx-vp9",
            "-b:v", maxrate,
            "-speed", spd,
            "-auto-alt-ref", "1",
            "-lag-in-frames", "25",
            "-frame-parallel", "0",
            "-row-mt", "1"
        ])

    cmd.extend([
        "-crf", str(crf),
    ])
    cmd.extend(shlex.split(oa))
    cmd.extend([
        "-pix_fmt", "yuv420p",
        "-max_muxing_queue_size", "99999",
        outfile
    ])

    scmd = " ".join(shlex.quote(x) for x in cmd)
    print(scmd, end="\n\n")

    p = sp.Popen(cmd, env=env)
    p.wait()
    if p.returncode == 0:
        status = 'OK'
    else:
        status = f'ERROR({p.returncode})'

    footer = f"\nx265ify.py {status}: original file last-modified {src_mod}, size {src_sz}\n"
    print(footer)

    os.rename(tmp_log, logfile)
    with open(logfile, 'ab+') as f:
        f.write(footer.encode('utf-8'))

    os.rmdir(lockdir)
    CLEANUP.remove(lockdir)
    return True


def cli_main():
    if len(sys.argv) < 2:
        print('drop one or more video files onto this python-file,')
        print('or see -h for help')
        sys.exit(1)

    # cfg: output quality; last value wins,
    #   anything above 20 has visible quality loss,
    #   48 is probably the highest you'll want for x265 (tiny files)
    #   52 is probably the highest you'll want for vp9
    _crf = [
        20,
        28,
        34,
        48,
        52,
        60,
        36,
    ]

    # cfg: time to spend, more time = better quality and smaller files,
    #   possible values from fastest to slowest:
    #   ultrafast superfast veryfast faster fast medium slow slower veryslow placebo
    # if vp9, format is vp9-s$speed-$max_bitrate
    # if x264, format is x264
    _preset = [
        "x264",
        "vp9-s1-3000k",
        "medium",
        "slow",
        "slower",
        "veryslow",
    ]

    # cfg: extra arguments on the input-side,
    #   -threads 4 lets FFmpeg hse 4 cores "generally"
    _inflags = [
        "-threads 4"
    ]

    # cfg: extra arguments on the output-side,
    # -r 30 sets output fps = 30,
    # x265 "ignores" -threads, pools=4 allocates 4 cores on one numa,
    #   https://x265.readthedocs.io/en/latest/cli.html#performance-options
    #   https://trac.ffmpeg.org/ticket/3730#comment:10
    # x265 -tune animation adds a lot of ringing
    _outflags = [
        "-tune animation -x265-params pools=4:vbv-maxrate=4000",
        "-tune animation -x265-params pools=4:vbv-maxrate=6000",
        "-x265-params pools=4:vbv-maxrate=7000:vbv-bufsize=14000",
        "-x265-params pools=4:vbv-maxrate=3000:vbv-bufsize=6000",
        "-x265-params pools=4:vbv-maxrate=3000:vbv-bufsize=6000 -r 30",
        "-x265-params vbv-maxrate=3000:vbv-bufsize=6000",
    ]

    # cfg: video filters to apply,
    #   mpdecimate is really good for mixed-fps sources (online music parties)
    _filters = [
        "pp7=3,vaguedenoiser=4",
        "pp7=3",
        "pp7=3,fftdnoiz=10:prev=1:next=1",
        "mpdecimate=frac=0.01,pp7=1,minterpolate=30000/1001:mci:mc_mode=aobmc:me_mode=bidir:vsbmc=1:mb_size=10",
        "copy",
        "pp=hb/vb/dr/al",
        "pp=ac,deflicker",
        "pp=ac,pp7=3",
        "pp=ac",
        "mpdecimate=frac=0.01",
    ]

    # cfg: resize video to WIDTH pixels 
    _width = [
        None,
        1280,
    ]

    # cfg: max fps in output
    _maxfps = [
        None,
        32,
    ]

    ap = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    ap.add_argument("-x", action="store_true", help="immediate exit (cli-mode)")
    ap.add_argument("-c", metavar="CRF", type=int, default=_crf[-1], help="quality level (higher=worse)")
    ap.add_argument("-p", metavar="PRESET", default=_preset[-1], help="preset")
    ap.add_argument("-ia", metavar="ARGS", default=_inflags[-1], help="input args")
    ap.add_argument("-oa", metavar="ARGS", default=_outflags[-1], help="output args")
    ap.add_argument("-vf", metavar="FILTERS", default=_filters[-1], help="video filters")
    ap.add_argument("-sz", metavar="RES", default=_width[-1], type=int, help="resize to NNN:-16")
    ap.add_argument("-r", metavar="FPS", default=_maxfps[-1], type=float, help="drop frames to hit this fps")
    ap.add_argument("files", metavar="FILEPATH", nargs="+", help="files to encode (or '-' to read list from stdin)")
    ar = ap.parse_args()

    name_extra = ''
    
    # hardsub support
    fc_path = os.path.join(tempfile.gettempdir(), 'ffhs-fonts.conf')
    with open(fc_path, 'wb') as f:
        f.write(r"""<?xml version="1.0"?>
<!DOCTYPE fontconfig SYSTEM "fonts.dtd">
<fontconfig><dir>.</dir></fontconfig>
""".encode('utf-8'))
    
    # lower piroity; will inherit  ['idle', 'below normal']
    if WINDOWS:
        sp.check_call(['wmic', 'process', 'where', f'processid={os.getpid()}', 'call', 'setpriority', 'idle'])
    else:
        os.nice(15)

    # ignore ctrl-c
    signal.signal(signal.SIGINT, sighandler)

    if ar.files == ['-']:
        ar.files = sys.stdin.read().strip().split('\n')

    arg_files = ar.files
    if WINDOWS:
        import glob

        arg_files = []
        for f in ar.files:
            if '*' in f:
                arg_files.extend(glob.glob(f))
            else:
                arg_files.append(f)

    files = []
    for n, src in enumerate(arg_files):
        if '://' in src:
            print(f"got URL #{n}: {src}")
            files.append(src)
            continue

        if not os.path.lexists(src):
            if src.startswith('*.') and len(src.split('.')) == 2:
                print(f"skipping nullglob [{src}]")
                continue

        if WINDOWS:
            short_src = src
            src = longify(src)
            print(f"got file #{n}: {short_src}\n  {src}\n")
        else:
            print(f"got file #{n}: {src}")

        if src is None or not os.path.lexists(src):
            if src and src.startswith('*.') and len(src.split('.')) == 2:
                print(f"skipping nullglob [{src}]")
                continue

            if WINDOWS:
                print(f'requesting shortname of [{src}]')
                need_shortnames()
            else:
                print(f"could not open file [{src}]")
            
            sys.exit(1)

        if not os.path.isfile(src):  # follows symlinks
            print("only files please, no folders")
            sys.exit(1)

        files.append(src)

    # support filenames starting with -
    files = [x if '://' in x else os.path.abspath(os.path.realpath(x)) for x in files]

    for n, src in enumerate(files):
        msg = f"x265ify [{n+1}/{len(files)}] {src}"
        set_title(msg)
        print(f"\n{msg}")
        try:
            if not transcode(ar, fc_path, src, name_extra):
                print("*** file skipped ***")
        except:
            set_title("x265ify FAILED")
            raise
        finally:
            try:
                os.rmdir(CLEANUP[0])
            except:
                pass
    
    set_title("x265ify done")


def gui_main():
    try:
        cli_main()
        sys.exit(0)
        input()
    except SystemExit as ex:
        if ex.code == 0:
            print('\n(finished, press ENTER to close)')
        else:
            print(f'\n(error {ex.code}, press ENTER to close)')
        
        tinput()
        raise
    except KeyboardInterrupt:
        pass
    except:
        traceback.print_exc()
        print('\n(crash, press ENTER to close)')
        tinput()
        sys.exit(1)


def main():
    for arg in sys.argv[1:]:
        if arg.startswith('-') and 'x' in arg:
            return cli_main()
    
    gui_main()


if __name__ == '__main__':
    main()


r"""
19:38
  slow: 50% cpu
slower: 39% 47% cpu (details/processes)
 vslow: 67% cpu?

c:\msys64\usr\bin\wget.exe -i urls.txt & c:\users\ed\dev\x265ify.py *.mp4 *.mkv *.ts
--limit-rate=5M

# for each file in srcdir, if there's a transcode in dstdir, ask if the src video should be deleted, and src.* + transcode-files should be moved into dstdir/done/
srcdir=/c/Users/ed/yt/; dstdir=/c/Users/ed/enc/; (cd "$srcdir" && find -maxdepth 1 -type f | cut -c3- | grep -E '\.(webm|mp4|mkv|ts)$') | while IFS= read -r src; do hits=("$dstdir$src".*); hit="${hits[0]}"; [ "$hit" = "${hit%\*}" ] || continue; printf '\n\033[1;33m%10d %s\033[0m\n' $(stat -c%s "$srcdir$src" || echo 0) "$srcdir$src"; printf '\033[1;33m%10d %s\033[0m\n' $(stat -c%s "$hit" || echo 0) "$hit"; ffprobe -i "$srcdir/$src" 2>&1 | grep Duration; ffprobe -i "$hit" 2>&1 | grep Duration; for pf in "$srcdir$src" "$hit"; do /c/Users/ed/bin/mpv.com -pause --fs=no --keep-open -ss 600 "$pf" </dev/null & done; wait; read -u1 -n1 -rp "delete src and move related+output to $dstdir/done/? [y/n/q] "; [ "$REPLY" = q ] && break; [ "$REPLY" = y ] || continue; rm -v "$srcdir/$src" && mv -nv "$hit"* "$dstdir/done/" && mv -nv "$srcdir${src%.*}".* "$dstdir/done/"; done
"""

